@@ -5,7 +5,6 @@ import enum
 import functools
 import inspect
 import socket
-import sys
 import warnings
 from textwrap import dedent
 from typing import (
@@ -17,30 +16,31 @@ from typing import (
     Iterable,
     Iterator,
     List,
+    Literal,
     Optional,
     Set,
     TypeVar,
     Union,
-    cast,
     overload,
 )
 
 import pytest
+from _pytest.mark.structures import get_unpacked_marks
 from pytest import (
+    Collector,
     Config,
     FixtureRequest,
     Function,
     Item,
+    Metafunc,
     Parser,
+    PytestCollectionWarning,
+    PytestDeprecationWarning,
     PytestPluginManager,
     Session,
+    StashKey,
 )
 
-if sys.version_info >= (3, 8):
-    from typing import Literal
-else:
-    from typing_extensions import Literal
-
 _R = TypeVar("_R")
 
 _ScopeName = Literal["session", "package", "module", "class", "function"]
@@ -60,6 +60,14 @@ FixtureDef = Any
 SubRequest = Any
 
 
+class PytestAsyncioError(Exception):
+    """Base class for exceptions raised by pytest-asyncio"""
+
+
+class MultipleEventLoopsRequestedError(PytestAsyncioError):
+    """Raised when a test requests multiple asyncio event loops."""
+
+
 class Mode(str, enum.Enum):
     AUTO = "auto"
     STRICT = "strict"
@@ -151,13 +159,8 @@ def _make_asyncio_fixture_function(obj: Any) -> None:
     obj._force_asyncio_fixture = True
 
 
-def _is_coroutine(obj: Any) -> bool:
-    """Check to see if an object is really an asyncio coroutine."""
-    return asyncio.iscoroutinefunction(obj)
-
-
 def _is_coroutine_or_asyncgen(obj: Any) -> bool:
-    return _is_coroutine(obj) or inspect.isasyncgenfunction(obj)
+    return asyncio.iscoroutinefunction(obj) or inspect.isasyncgenfunction(obj)
 
 
 def _get_asyncio_mode(config: Config) -> Mode:
@@ -181,6 +184,12 @@ def pytest_configure(config: Config) -> None:
         "mark the test as a coroutine, it will be "
         "run using an asyncio event loop",
     )
+    config.addinivalue_line(
+        "markers",
+        "asyncio_event_loop: "
+        "Provides an asyncio event loop in the scope of the marked test "
+        "class or module",
+    )
 
 
 @pytest.hookimpl(tryfirst=True)
@@ -191,11 +200,17 @@ def pytest_report_header(config: Config) -> List[str]:
 
 
 def _preprocess_async_fixtures(
-    config: Config,
+    collector: Collector,
     processed_fixturedefs: Set[FixtureDef],
 ) -> None:
+    config = collector.config
     asyncio_mode = _get_asyncio_mode(config)
     fixturemanager = config.pluginmanager.get_plugin("funcmanage")
+    event_loop_fixture_id = "event_loop"
+    for node, mark in collector.iter_markers_with_node("asyncio_event_loop"):
+        event_loop_fixture_id = node.stash.get(_event_loop_fixture_id, None)
+        if event_loop_fixture_id:
+            break
     for fixtures in fixturemanager._arg2fixturedefs.values():
         for fixturedef in fixtures:
             func = fixturedef.func
@@ -208,37 +223,52 @@ def _preprocess_async_fixtures(
                 # This applies to pytest_trio fixtures, for example
                 continue
             _make_asyncio_fixture_function(func)
-            _inject_fixture_argnames(fixturedef)
-            _synchronize_async_fixture(fixturedef)
+            function_signature = inspect.signature(func)
+            if "event_loop" in function_signature.parameters:
+                warnings.warn(
+                    PytestDeprecationWarning(
+                        f"{func.__name__} is asynchronous and explicitly "
+                        f'requests the "event_loop" fixture. Asynchronous fixtures and '
+                        f'test functions should use "asyncio.get_running_loop()" '
+                        f"instead."
+                    )
+                )
+            _inject_fixture_argnames(fixturedef, event_loop_fixture_id)
+            _synchronize_async_fixture(fixturedef, event_loop_fixture_id)
             assert _is_asyncio_fixture_function(fixturedef.func)
             processed_fixturedefs.add(fixturedef)
 
 
-def _inject_fixture_argnames(fixturedef: FixtureDef) -> None:
+def _inject_fixture_argnames(
+    fixturedef: FixtureDef, event_loop_fixture_id: str
+) -> None:
     """
     Ensures that `request` and `event_loop` are arguments of the specified fixture.
     """
     to_add = []
-    for name in ("request", "event_loop"):
+    for name in ("request", event_loop_fixture_id):
         if name not in fixturedef.argnames:
             to_add.append(name)
     if to_add:
         fixturedef.argnames += tuple(to_add)
 
 
-def _synchronize_async_fixture(fixturedef: FixtureDef) -> None:
+def _synchronize_async_fixture(
+    fixturedef: FixtureDef, event_loop_fixture_id: str
+) -> None:
     """
     Wraps the fixture function of an async fixture in a synchronous function.
     """
     if inspect.isasyncgenfunction(fixturedef.func):
-        _wrap_asyncgen_fixture(fixturedef)
+        _wrap_asyncgen_fixture(fixturedef, event_loop_fixture_id)
     elif inspect.iscoroutinefunction(fixturedef.func):
-        _wrap_async_fixture(fixturedef)
+        _wrap_async_fixture(fixturedef, event_loop_fixture_id)
 
 
 def _add_kwargs(
     func: Callable[..., Any],
     kwargs: Dict[str, Any],
+    event_loop_fixture_id: str,
     event_loop: asyncio.AbstractEventLoop,
     request: SubRequest,
 ) -> Dict[str, Any]:
@@ -246,8 +276,8 @@ def _add_kwargs(
     ret = kwargs.copy()
     if "request" in sig.parameters:
         ret["request"] = request
-    if "event_loop" in sig.parameters:
-        ret["event_loop"] = event_loop
+    if event_loop_fixture_id in sig.parameters:
+        ret[event_loop_fixture_id] = event_loop
     return ret
 
 
@@ -270,17 +300,18 @@ def _perhaps_rebind_fixture_func(
     return func
 
 
-def _wrap_asyncgen_fixture(fixturedef: FixtureDef) -> None:
+def _wrap_asyncgen_fixture(fixturedef: FixtureDef, event_loop_fixture_id: str) -> None:
     fixture = fixturedef.func
 
     @functools.wraps(fixture)
-    def _asyncgen_fixture_wrapper(
-        event_loop: asyncio.AbstractEventLoop, request: SubRequest, **kwargs: Any
-    ):
+    def _asyncgen_fixture_wrapper(request: SubRequest, **kwargs: Any):
         func = _perhaps_rebind_fixture_func(
             fixture, request.instance, fixturedef.unittest
         )
-        gen_obj = func(**_add_kwargs(func, kwargs, event_loop, request))
+        event_loop = kwargs.pop(event_loop_fixture_id)
+        gen_obj = func(
+            **_add_kwargs(func, kwargs, event_loop_fixture_id, event_loop, request)
+        )
 
         async def setup():
             res = await gen_obj.__anext__()
@@ -308,19 +339,20 @@ def _wrap_asyncgen_fixture(fixturedef: FixtureDef) -> None:
     fixturedef.func = _asyncgen_fixture_wrapper
 
 
-def _wrap_async_fixture(fixturedef: FixtureDef) -> None:
+def _wrap_async_fixture(fixturedef: FixtureDef, event_loop_fixture_id: str) -> None:
     fixture = fixturedef.func
 
     @functools.wraps(fixture)
-    def _async_fixture_wrapper(
-        event_loop: asyncio.AbstractEventLoop, request: SubRequest, **kwargs: Any
-    ):
+    def _async_fixture_wrapper(request: SubRequest, **kwargs: Any):
         func = _perhaps_rebind_fixture_func(
             fixture, request.instance, fixturedef.unittest
         )
+        event_loop = kwargs.pop(event_loop_fixture_id)
 
         async def setup():
-            res = await func(**_add_kwargs(func, kwargs, event_loop, request))
+            res = await func(
+                **_add_kwargs(func, kwargs, event_loop_fixture_id, event_loop, request)
+            )
             return res
 
         return event_loop.run_until_complete(setup())
@@ -328,11 +360,144 @@ def _wrap_async_fixture(fixturedef: FixtureDef) -> None:
     fixturedef.func = _async_fixture_wrapper
 
 
+class PytestAsyncioFunction(Function):
+    """Base class for all test functions managed by pytest-asyncio."""
+
+    @classmethod
+    def substitute(cls, item: Function, /) -> Function:
+        """
+        Returns a PytestAsyncioFunction if there is an implementation that can handle
+        the specified function item.
+
+        If no implementation of PytestAsyncioFunction can handle the specified item,
+        the item is returned unchanged.
+        """
+        for subclass in cls.__subclasses__():
+            if subclass._can_substitute(item):
+                return subclass._from_function(item)
+        return item
+
+    @classmethod
+    def _from_function(cls, function: Function, /) -> Function:
+        """
+        Instantiates this specific PytestAsyncioFunction type from the specified
+        Function item.
+        """
+        subclass_instance = cls.from_parent(
+            function.parent,
+            name=function.name,
+            callspec=getattr(function, "callspec", None),
+            callobj=function.obj,
+            fixtureinfo=function._fixtureinfo,
+            keywords=function.keywords,
+            originalname=function.originalname,
+        )
+        subclassed_function_signature = inspect.signature(subclass_instance.obj)
+        if "event_loop" in subclassed_function_signature.parameters:
+            subclass_instance.warn(
+                PytestDeprecationWarning(
+                    f"{subclass_instance.name} is asynchronous and explicitly "
+                    f'requests the "event_loop" fixture. Asynchronous fixtures and '
+                    f'test functions should use "asyncio.get_running_loop()" instead.'
+                )
+            )
+        return subclass_instance
+
+    @staticmethod
+    def _can_substitute(item: Function) -> bool:
+        """Returns whether the specified function can be replaced by this class"""
+        raise NotImplementedError()
+
+
+class Coroutine(PytestAsyncioFunction):
+    """Pytest item created by a coroutine"""
+
+    @staticmethod
+    def _can_substitute(item: Function) -> bool:
+        func = item.obj
+        return asyncio.iscoroutinefunction(func)
+
+    def runtest(self) -> None:
+        if self.get_closest_marker("asyncio"):
+            self.obj = wrap_in_sync(
+                # https://github.com/pytest-dev/pytest-asyncio/issues/596
+                self.obj,  # type: ignore[has-type]
+            )
+        super().runtest()
+
+
+class AsyncGenerator(PytestAsyncioFunction):
+    """Pytest item created by an asynchronous generator"""
+
+    @staticmethod
+    def _can_substitute(item: Function) -> bool:
+        func = item.obj
+        return inspect.isasyncgenfunction(func)
+
+    @classmethod
+    def _from_function(cls, function: Function, /) -> Function:
+        async_gen_item = super()._from_function(function)
+        unsupported_item_type_message = (
+            f"Tests based on asynchronous generators are not supported. "
+            f"{function.name} will be ignored."
+        )
+        async_gen_item.warn(PytestCollectionWarning(unsupported_item_type_message))
+        async_gen_item.add_marker(
+            pytest.mark.xfail(run=False, reason=unsupported_item_type_message)
+        )
+        return async_gen_item
+
+
+class AsyncStaticMethod(PytestAsyncioFunction):
+    """
+    Pytest item that is a coroutine or an asynchronous generator
+    decorated with staticmethod
+    """
+
+    @staticmethod
+    def _can_substitute(item: Function) -> bool:
+        func = item.obj
+        return isinstance(func, staticmethod) and _is_coroutine_or_asyncgen(
+            func.__func__
+        )
+
+    def runtest(self) -> None:
+        if self.get_closest_marker("asyncio"):
+            self.obj = wrap_in_sync(
+                # https://github.com/pytest-dev/pytest-asyncio/issues/596
+                self.obj,  # type: ignore[has-type]
+            )
+        super().runtest()
+
+
+class AsyncHypothesisTest(PytestAsyncioFunction):
+    """
+    Pytest item that is coroutine or an asynchronous generator decorated by
+    @hypothesis.given.
+    """
+
+    @staticmethod
+    def _can_substitute(item: Function) -> bool:
+        func = item.obj
+        return getattr(
+            func, "is_hypothesis_test", False
+        ) and asyncio.iscoroutinefunction(func.hypothesis.inner_test)
+
+    def runtest(self) -> None:
+        if self.get_closest_marker("asyncio"):
+            self.obj.hypothesis.inner_test = wrap_in_sync(
+                self.obj.hypothesis.inner_test,
+            )
+        super().runtest()
+
+
 _HOLDER: Set[FixtureDef] = set()
 
 
-@pytest.hookimpl(tryfirst=True)
-def pytest_pycollect_makeitem(
+# The function name needs to start with "pytest_"
+# see https://github.com/pytest-dev/pytest/issues/11307
+@pytest.hookimpl(specname="pytest_pycollect_makeitem", tryfirst=True)
+def pytest_pycollect_makeitem_preprocess_async_fixtures(
     collector: Union[pytest.Module, pytest.Class], name: str, obj: object
 ) -> Union[
     pytest.Item, pytest.Collector, List[Union[pytest.Item, pytest.Collector]], None
@@ -340,10 +505,100 @@ def pytest_pycollect_makeitem(
     """A pytest hook to collect asyncio coroutines."""
     if not collector.funcnamefilter(name):
         return None
-    _preprocess_async_fixtures(collector.config, _HOLDER)
+    _preprocess_async_fixtures(collector, _HOLDER)
     return None
 
 
+# The function name needs to start with "pytest_"
+# see https://github.com/pytest-dev/pytest/issues/11307
+@pytest.hookimpl(specname="pytest_pycollect_makeitem", hookwrapper=True)
+def pytest_pycollect_makeitem_convert_async_functions_to_subclass(
+    collector: Union[pytest.Module, pytest.Class], name: str, obj: object
+) -> Union[
+    pytest.Item, pytest.Collector, List[Union[pytest.Item, pytest.Collector]], None
+]:
+    """
+    Converts coroutines and async generators collected as pytest.Functions
+    to AsyncFunction items.
+    """
+    hook_result = yield
+    node_or_list_of_nodes = hook_result.get_result()
+    if not node_or_list_of_nodes:
+        return
+    try:
+        node_iterator = iter(node_or_list_of_nodes)
+    except TypeError:
+        # Treat single node as a single-element iterable
+        node_iterator = iter((node_or_list_of_nodes,))
+    updated_node_collection = []
+    for node in node_iterator:
+        updated_item = node
+        if isinstance(node, Function):
+            updated_item = PytestAsyncioFunction.substitute(node)
+        updated_node_collection.append(updated_item)
+
+    hook_result.force_result(updated_node_collection)
+
+
+_event_loop_fixture_id = StashKey[str]
+
+
+@pytest.hookimpl
+def pytest_collectstart(collector: pytest.Collector):
+    if not isinstance(collector, (pytest.Class, pytest.Module)):
+        return
+    # pytest.Collector.own_markers is empty at this point,
+    # so we rely on _pytest.mark.structures.get_unpacked_marks
+    marks = get_unpacked_marks(collector.obj, consider_mro=True)
+    for mark in marks:
+        if not mark.name == "asyncio_event_loop":
+            continue
+        event_loop_policy = mark.kwargs.get("policy", asyncio.get_event_loop_policy())
+        policy_params = (
+            event_loop_policy
+            if isinstance(event_loop_policy, Iterable)
+            else (event_loop_policy,)
+        )
+
+        # There seem to be issues when a fixture is shadowed by another fixture
+        # and both differ in their params.
+        # https://github.com/pytest-dev/pytest/issues/2043
+        # https://github.com/pytest-dev/pytest/issues/11350
+        # As such, we assign a unique name for each event_loop fixture.
+        # The fixture name is stored in the collector's Stash, so it can
+        # be injected when setting up the test
+        event_loop_fixture_id = f"{collector.nodeid}::<event_loop>"
+        collector.stash[_event_loop_fixture_id] = event_loop_fixture_id
+
+        @pytest.fixture(
+            scope="class" if isinstance(collector, pytest.Class) else "module",
+            name=event_loop_fixture_id,
+            params=policy_params,
+            ids=tuple(type(policy).__name__ for policy in policy_params),
+        )
+        def scoped_event_loop(
+            *args,  # Function needs to accept "cls" when collected by pytest.Class
+            request,
+        ) -> Iterator[asyncio.AbstractEventLoop]:
+            new_loop_policy = request.param
+            old_loop_policy = asyncio.get_event_loop_policy()
+            old_loop = asyncio.get_event_loop()
+            asyncio.set_event_loop_policy(new_loop_policy)
+            loop = asyncio.new_event_loop()
+            asyncio.set_event_loop(loop)
+            yield loop
+            loop.close()
+            asyncio.set_event_loop_policy(old_loop_policy)
+            asyncio.set_event_loop(old_loop)
+
+        # @pytest.fixture does not register the fixture anywhere, so pytest doesn't
+        # know it exists. We work around this by attaching the fixture function to the
+        # collected Python class, where it will be picked up by pytest.Class.collect()
+        # or pytest.Module.collect(), respectively
+        collector.obj.__pytest_asyncio_scoped_event_loop = scoped_event_loop
+        break
+
+
 def pytest_collection_modifyitems(
     session: Session, config: Config, items: List[Item]
 ) -> None:
@@ -352,29 +607,58 @@ def pytest_collection_modifyitems(
 
     The mark is only applied in `AUTO` mode. It is applied to:
 
-      - coroutines
-      - staticmethods wrapping coroutines
+      - coroutines and async generators
       - Hypothesis tests wrapping coroutines
+      - staticmethods wrapping coroutines
 
     """
     if _get_asyncio_mode(config) != Mode.AUTO:
         return
-    function_items = (item for item in items if isinstance(item, Function))
-    for function_item in function_items:
-        function = function_item.obj
-        if isinstance(function, staticmethod):
-            # staticmethods need to be unwrapped.
-            function = function.__func__
-        if (
-            _is_coroutine(function)
-            or _is_hypothesis_test(function)
-            and _hypothesis_test_wraps_coroutine(function)
-        ):
-            function_item.add_marker("asyncio")
+    for item in items:
+        if isinstance(item, PytestAsyncioFunction):
+            item.add_marker("asyncio")
 
 
-def _hypothesis_test_wraps_coroutine(function: Any) -> bool:
-    return _is_coroutine(function.hypothesis.inner_test)
+_REDEFINED_EVENT_LOOP_FIXTURE_WARNING = dedent(
+    """\
+    The event_loop fixture provided by pytest-asyncio has been redefined in
+    %s:%d
+    Replacing the event_loop fixture with a custom implementation is deprecated
+    and will lead to errors in the future.
+    If you want to request an asyncio event loop with a class or module scope,
+    please attach the asyncio_event_loop mark to the respective class or module.
+    """
+)
+
+
+@pytest.hookimpl(tryfirst=True)
+def pytest_generate_tests(metafunc: Metafunc) -> None:
+    for event_loop_provider_node, _ in metafunc.definition.iter_markers_with_node(
+        "asyncio_event_loop"
+    ):
+        event_loop_fixture_id = event_loop_provider_node.stash.get(
+            _event_loop_fixture_id, None
+        )
+        if event_loop_fixture_id:
+            # This specific fixture name may already be in metafunc.argnames, if this
+            # test indirectly depends on the fixture. For example, this is the case
+            # when the test depends on an async fixture, both of which share the same
+            # asyncio_event_loop mark.
+            if event_loop_fixture_id in metafunc.fixturenames:
+                continue
+            fixturemanager = metafunc.config.pluginmanager.get_plugin("funcmanage")
+            if "event_loop" in metafunc.fixturenames:
+                raise MultipleEventLoopsRequestedError(
+                    _MULTIPLE_LOOPS_REQUESTED_ERROR
+                    % (metafunc.definition.nodeid, event_loop_provider_node.nodeid),
+                )
+            # Add the scoped event loop fixture to Metafunc's list of fixture names and
+            # fixturedefs and leave the actual parametrization to pytest
+            metafunc.fixturenames.insert(0, event_loop_fixture_id)
+            metafunc._arg2fixturedefs[
+                event_loop_fixture_id
+            ] = fixturemanager._arg2fixturedefs[event_loop_fixture_id]
+            break
 
 
 @pytest.hookimpl(hookwrapper=True)
@@ -395,6 +679,16 @@ def pytest_fixture_setup(
         )
         outcome = yield
         loop = outcome.get_result()
+        # Weird behavior was observed when checking for an attribute of FixtureDef.func
+        # Instead, we now check for a special attribute of the returned event loop
+        fixture_filename = inspect.getsourcefile(fixturedef.func)
+        if not getattr(loop, "__original_fixture_loop", False):
+            _, fixture_line_number = inspect.getsourcelines(fixturedef.func)
+            warnings.warn(
+                _REDEFINED_EVENT_LOOP_FIXTURE_WARNING
+                % (fixture_filename, fixture_line_number),
+                DeprecationWarning,
+            )
         policy = asyncio.get_event_loop_policy()
         try:
             with warnings.catch_warnings():
@@ -468,40 +762,31 @@ def _provide_clean_event_loop() -> None:
 
 
 @pytest.hookimpl(tryfirst=True, hookwrapper=True)
-def pytest_pyfunc_call(pyfuncitem: pytest.Function) -> Optional[object]:
+def pytest_pyfunc_call(pyfuncitem: Function) -> Optional[object]:
     """
     Pytest hook called before a test case is run.
 
     Wraps marked tests in a synchronous function
     where the wrapped test coroutine is executed in an event loop.
     """
-    marker = pyfuncitem.get_closest_marker("asyncio")
-    if marker is not None:
-        funcargs: Dict[str, object] = pyfuncitem.funcargs  # type: ignore[name-defined]
-        loop = cast(asyncio.AbstractEventLoop, funcargs["event_loop"])
-        if _is_hypothesis_test(pyfuncitem.obj):
-            pyfuncitem.obj.hypothesis.inner_test = wrap_in_sync(
-                pyfuncitem,
-                pyfuncitem.obj.hypothesis.inner_test,
-                _loop=loop,
-            )
+    if pyfuncitem.get_closest_marker("asyncio") is not None:
+        if isinstance(pyfuncitem, PytestAsyncioFunction):
+            pass
         else:
-            pyfuncitem.obj = wrap_in_sync(
-                pyfuncitem,
-                pyfuncitem.obj,
-                _loop=loop,
+            pyfuncitem.warn(
+                pytest.PytestWarning(
+                    f"The test {pyfuncitem} is marked with '@pytest.mark.asyncio' "
+                    "but it is not an async function. "
+                    "Please remove the asyncio mark. "
+                    "If the test is not marked explicitly, "
+                    "check for global marks applied via 'pytestmark'."
+                )
             )
     yield
 
 
-def _is_hypothesis_test(function: Any) -> bool:
-    return getattr(function, "is_hypothesis_test", False)
-
-
 def wrap_in_sync(
-    pyfuncitem: pytest.Function,
     func: Callable[..., Awaitable[Any]],
-    _loop: asyncio.AbstractEventLoop,
 ):
     """Return a sync wrapper around an async function executing it in the
     current event loop."""
@@ -516,17 +801,7 @@ def wrap_in_sync(
     @functools.wraps(func)
     def inner(*args, **kwargs):
         coro = func(*args, **kwargs)
-        if not inspect.isawaitable(coro):
-            pyfuncitem.warn(
-                pytest.PytestWarning(
-                    f"The test {pyfuncitem} is marked with '@pytest.mark.asyncio' "
-                    "but it is not an async function. "
-                    "Please remove asyncio marker. "
-                    "If the test is not marked explicitly, "
-                    "check for global markers applied via 'pytestmark'."
-                )
-            )
-            return
+        _loop = asyncio.get_event_loop()
         task = asyncio.ensure_future(coro, loop=_loop)
         try:
             _loop.run_until_complete(task)
@@ -542,15 +817,32 @@ def wrap_in_sync(
     return inner
 
 
+_MULTIPLE_LOOPS_REQUESTED_ERROR = dedent(
+    """\
+        Multiple asyncio event loops with different scopes have been requested
+        by %s. The test explicitly requests the event_loop fixture, while another
+        event loop is provided by %s.
+        Remove "event_loop" from the requested fixture in your test to run the test
+        in a larger-scoped event loop or remove the "asyncio_event_loop" mark to run
+        the test in a function-scoped event loop.
+    """
+)
+
+
 def pytest_runtest_setup(item: pytest.Item) -> None:
     marker = item.get_closest_marker("asyncio")
     if marker is None:
         return
+    event_loop_fixture_id = "event_loop"
+    for node, mark in item.iter_markers_with_node("asyncio_event_loop"):
+        event_loop_fixture_id = node.stash.get(_event_loop_fixture_id, None)
+        if event_loop_fixture_id:
+            break
     fixturenames = item.fixturenames  # type: ignore[attr-defined]
     # inject an event loop fixture for all async tests
     if "event_loop" in fixturenames:
         fixturenames.remove("event_loop")
-    fixturenames.insert(0, "event_loop")
+    fixturenames.insert(0, event_loop_fixture_id)
     obj = getattr(item, "obj", None)
     if not getattr(obj, "hypothesis", False) and getattr(
         obj, "is_hypothesis_test", False
@@ -565,6 +857,12 @@ def pytest_runtest_setup(item: pytest.Item) -> None:
 def event_loop(request: FixtureRequest) -> Iterator[asyncio.AbstractEventLoop]:
     """Create an instance of the default event loop for each test case."""
     loop = asyncio.get_event_loop_policy().new_event_loop()
+    # Add a magic value to the event loop, so pytest-asyncio can determine if the
+    # event_loop fixture was overridden. Other implementations of event_loop don't
+    # set this value.
+    # The magic value must be set as part of the function definition, because pytest
+    # seems to have multiple instances of the same FixtureDef or fixture function
+    loop.__original_fixture_loop = True  # type: ignore[attr-defined]
     yield loop
     loop.close()
 
